from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, Any


if TYPE_CHECKING:
    from collections.abc import Callable


def cache_with_transforms[**P, R](
    *,
    arg_transformers: dict[int, Callable[[Any], Any]] | None = None,
    kwarg_transformers: dict[str, Callable[[Any], Any]] | None = None,
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """A caching decorator with transformation functions for args and kwargs.

    Can be used to make specific args / kwargs hashable.
    Also adds cache and cache_info objects to the decorated function.

    Args:
        arg_transformers: Dict mapping positional args indices to transformer functions
        kwarg_transformers: Dict mapping kwargs names to transformer functions

    Returns:
        A decorator function that caches results based on transformed arguments
    """
    arg_transformers = arg_transformers or {}
    kwarg_transformers = kwarg_transformers or {}

    def decorator(func: Callable[P, R]) -> Callable[P, R]:
        cache: dict[tuple[Any, ...], R] = {}

        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            # Transform positional arguments
            transformed_args = tuple(
                arg_transformers.get(i, lambda x: x)(arg) for i, arg in enumerate(args)
            )

            # Transform keyword arguments
            transformed_kwargs = {
                key: kwarg_transformers.get(key, lambda x: x)(value)
                for key, value in sorted(kwargs.items())
            }

            # Create cache key from transformed arguments
            cache_key = (transformed_args, tuple(transformed_kwargs.items()))

            if cache_key not in cache:
                cache[cache_key] = func(*args, **kwargs)
            return cache[cache_key]

        def cache_info() -> dict[str, int]:
            """Return information about cache hits and size."""
            return {"cache_size": len(cache)}

        wrapper.cache_info = cache_info  # type: ignore
        wrapper.cache = cache  # type: ignore

        return wrapper

    return decorator


if __name__ == "__main__":
    import upath

    @cache_with_transforms(arg_transformers={0: lambda p: upath.UPath(p).resolve()})
    def read_file_content(filepath: str | upath.UPath) -> str:
        """Read and return the content of a file."""
        with upath.UPath(filepath).open() as f:
            return f.read()

    # These calls will use the same cache entry
    content1 = read_file_content("pyproject.toml")
    content1 = read_file_content("mkdocs.yml")
    content2 = read_file_content(upath.UPath("pyproject.toml"))
    content3 = read_file_content(upath.UPath("./pyproject.toml").absolute())

    # Check cache statistics
    print(read_file_content.cache_info())  # type: ignore



"""Filesystem implementation for jinja environment templates."""

from __future__ import annotations

import logging
from typing import Any, Literal, override

import fsspec
from fsspec.implementations import memory
import jinja2


logger = logging.getLogger(__name__)


class JinjaLoaderFileSystem(fsspec.AbstractFileSystem):
    """A FsSpec Filesystem implementation for jinja environment templates."""

    protocol = "jinja"
    async_impl = True

    def __init__(self, env: jinja2.Environment) -> None:
        """Initialize a JinjaLoader filesystem.

        Args:
            env: The environment of the loaders to get a filesystem for.
        """
        super().__init__()
        self.env = env

    @override
    def isdir(self, path: str) -> bool:
        """Check if path is a directory.

        Args:
            path: Path to check

        Returns:
            True if path is a directory
        """
        from upath import UPath

        if not self.env.loader:
            return False

        clean_path = UPath(path).as_posix().strip("/")
        if clean_path in {"", "/", "."}:
            return True

        templates = self.env.loader.list_templates()
        return any(template.startswith(f"{clean_path}/") for template in templates)

    @override
    def isfile(self, path: str) -> bool:
        """Check if path is a file.

        Args:
            path: Path to check

        Returns:
            True if path is a file
        """
        if not self.env.loader:
            return False

        try:
            self.env.loader.get_source(self.env, path)
        except jinja2.TemplateNotFound:
            return False
        else:
            return True

    @override
    def cat_file(self, path: str, **kwargs: Any) -> bytes:
        """Get contents of file as bytes.

        Args:
            path: Path to file
            **kwargs: Additional arguments (unused)

        Returns:
            File contents as bytes

        Raises:
            FileNotFoundError: If template not found or env has no loader
        """
        if not self.env.loader:
            msg = "Environment has no loader"
            raise FileNotFoundError(msg)

        try:
            source, _, _ = self.env.loader.get_source(self.env, path)
            return source.encode()
        except jinja2.TemplateNotFound as exc:
            msg = f"Template not found: {path}"
            raise FileNotFoundError(msg) from exc

    @override
    def cat(self, path: str | list[str], **kwargs: Any) -> bytes | dict[str, bytes]:
        """Get contents of file(s).

        Args:
            path: Path or list of paths
            **kwargs: Additional arguments

        Returns:
            File contents as bytes or dict of path -> contents
        """
        if isinstance(path, str):
            return self.cat_file(path, **kwargs)

        return {p: self.cat_file(p, **kwargs) for p in path}

    async def _cat_file(self, path: str, **kwargs: Any) -> bytes:
        """Async implementation of cat_file.

        Args:
            path: Path to file
            **kwargs: Additional arguments

        Returns:
            File contents as bytes
        """
        return self.cat_file(path, **kwargs)

    @override
    def ls(self, path: str, detail: bool = True, **kwargs: Any) -> list[dict[str, str]] | list[str]:
        """List contents of path.

        Args:
            path: The path to list
            detail: If True, return a list of dictionaries, else return a list of paths
            **kwargs: Additional arguments (unused)

        Returns:
            List of paths or file details

        Raises:
            FileNotFoundError: If path doesn't exist or env has no loader
        """
        from upath import UPath

        if not self.env.loader:
            msg = "Environment has no loader"
            raise FileNotFoundError(msg)

        templates = self.env.loader.list_templates()
        clean_path = UPath(path).as_posix().strip("/")

        if clean_path in {"", "/", "."}:
            return self._list_root(templates, detail)
        return self._list_subdirectory(templates, clean_path, detail)

    def _list_root(self, templates: list[str], detail: bool) -> list[dict[str, str]] | list[str]:
        """List contents of root directory."""
        root_files = [path for path in templates if "/" not in path]
        root_dirs = {
            path.split("/")[0] for path in templates if "/" in path and path not in root_files
        }

        if detail:
            file_entries = [{"name": path, "type": "file"} for path in root_files]
            dir_entries = [{"name": path, "type": "directory"} for path in root_dirs]
            return dir_entries + file_entries
        return list(root_dirs) + root_files

    def _list_subdirectory(
        self, templates: list[str], path: str, detail: bool
    ) -> list[dict[str, str]] | list[str]:
        """List contents of a subdirectory."""
        # Get all templates that start with the path
        relevant_templates = [template for template in templates if template.startswith(f"{path}/")]

        if not relevant_templates:
            msg = f"Directory not found: {path}"
            raise FileNotFoundError(msg)

        # Get immediate children only
        items: set[str] = set()
        for template in relevant_templates:
            # Remove the path prefix and split the remaining path
            relative_path = template[len(f"{path}/") :].split("/", 1)[0]
            items.add(relative_path)

        # Sort for consistent ordering
        sorted_items = sorted(items)

        if detail:
            return [
                {
                    "name": item,
                    # If there's no extension or if it appears in the full paths
                    # with something after it, it's a directory
                    "type": "directory"
                    if ("." not in item or any(t.startswith(f"{path}/{item}/") for t in templates))
                    else "file",
                }
                for item in sorted_items
            ]
        return list(sorted_items)

    async def _ls(
        self, path: str, detail: bool = True, **kwargs: Any
    ) -> list[dict[str, str]] | list[str]:
        """Async implementation of ls."""
        return self.ls(path, detail=detail, **kwargs)

    @override
    def _open(
        self,
        path: str,
        mode: Literal["rb", "wb", "ab"] = "rb",
        block_size: int | None = None,
        autocommit: bool = True,
        cache_options: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> memory.MemoryFile:
        """Open a file."""
        if not self.env.loader:
            msg = "Environment has no loader"
            raise FileNotFoundError(msg)

        try:
            source, _, _ = self.env.loader.get_source(self.env, path)
            return memory.MemoryFile(fs=self, path=path, data=source.encode())
        except jinja2.TemplateNotFound as exc:
            msg = f"Template not found: {path}"
            raise FileNotFoundError(msg) from exc

    async def _open_async(
        self,
        path: str,
        mode: Literal["rb", "wb", "ab"] = "rb",
        block_size: int | None = None,
        autocommit: bool = True,
        cache_options: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> memory.MemoryFile:
        """Async implementation of _open."""
        return self._open(
            path,
            mode=mode,
            block_size=block_size,
            autocommit=autocommit,
            cache_options=cache_options,
            **kwargs,
        )

    @override
    def exists(self, path: str, **_kwargs: Any) -> bool:
        """Check if path exists.

        Args:
            path: Path to check

        Returns:
            True if path exists as a file or directory
        """
        return self.isfile(path) or self.isdir(path)

    @override
    def info(self, path: str, **kwargs: Any) -> dict[str, Any]:
        """Get info about a path.

        Args:
            path: Path to get info for
            **kwargs: Additional arguments (unused)

        Returns:
            Dictionary of path information

        Raises:
            FileNotFoundError: If path doesn't exist
        """
        if self.isfile(path):
            content = self.cat_file(path)
            return {
                "name": path,
                "size": len(content),
                "type": "file",
                "created": None,  # Jinja doesn't track these
                "modified": None,
            }
        if self.isdir(path):
            return {
                "name": path,
                "size": 0,
                "type": "directory",
                "created": None,
                "modified": None,
            }

        msg = f"Path not found: {path}"
        raise FileNotFoundError(msg)


if __name__ == "__main__":
    from jinjarope import loaders

    fsspec.register_implementation("jinja", JinjaLoaderFileSystem)
    template_env = jinja2.Environment(loader=loaders.PackageLoader("jinjarope"))
    filesystem = fsspec.filesystem("jinja", env=template_env)
    print(filesystem.ls(""))



from __future__ import annotations

import os
from typing import Any

from dotenv import load_dotenv

# import litellm
from jinjarope import lazylitellm


load_dotenv()


litellm = lazylitellm.LazyLiteLLM()


def llm_generate_image(
    prompt: str,
    model: str | None = None,
    token: str | None = None,
    base_url: str | None = None,
    size: str = "1024x1024",
    quality: str = "standard",
    as_b64_json: bool = False,
    **kwargs: Any,
) -> str | None:
    """Generate an image using the LLM API and returns the URL.

    Args:
        prompt: The prompt to generate an image from.
        model: The model to use. Defaults to None.
        token: The API token. Defaults to None.
        base_url: The base URL of the API. Defaults to None.
        size: The size of the generated image. Defaults to "1024x1024".
        quality: The quality of the generated image. Defaults to "standard".
        as_b64_json: Return b64-encoded image instead of URL.
        kwargs: Additional keyword arguments passed to litellm.image_generation.

    Returns:
        The generated image response.
    """
    response = litellm.image_generation(
        prompt=prompt,
        model=model or os.getenv("OPENAI_IMAGE_MODEL"),
        api_key=token or os.getenv("OPENAI_API_TOKEN"),
        api_base=base_url,
        size=size,
        quality=quality,
        response_format="b64_json" if as_b64_json else "url",
        **kwargs,
    )
    # Check if the API result is valid
    if response and response.data and len(response.data) > 0:
        # TODO: <img src="data:image/png;base64,iVBORw0KG..." />
        return response.data[0].b64_json if as_b64_json else response.data[0].url
    return None


if __name__ == "__main__":
    response = llm_generate_image(
        prompt="A dog!",
        model="ollama/llava",
    )
    print(response)
    print(response)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

from jinjarope import deepmerge


if TYPE_CHECKING:
    from collections.abc import Callable


SerializeFormatStr = Literal["yaml", "json", "ini", "toml"]


def serialize(data: Any, fmt: SerializeFormatStr, **kwargs: Any) -> str:
    """Serialize given json-like object to given format.

    Args:
        data: The data to serialize
        fmt: The serialization format
        kwargs: Keyword arguments passed to the dumper function
    """
    from yamling import dump_universal

    return dump_universal.dump(data, mode=fmt, **kwargs)


def deserialize(data: str, fmt: SerializeFormatStr, **kwargs: Any) -> Any:
    """Serialize given json-like object to given format.

    Args:
        data: The data to deserialize
        fmt: The serialization format
        kwargs: Keyword arguments passed to the loader function
    """
    from yamling import load_universal

    return load_universal.load(data, mode=fmt, **kwargs)


def dig(
    data: dict[str, Any] | list[Any] | str,
    *sections: str,
    keep_path: bool = False,
    dig_yaml_lists: bool = True,
) -> Any:
    """Try to get data with given section path from a dict-list structure.

    If a list is encountered and dig_yaml_lists is true, treat it like a list of
    {"identifier", {subdict}} items, as used in MkDocs config for
    plugins & extensions.
    If Key path does not exist, return None.

    Args:
        data: The data to dig into
        sections: Sections to dig into
        keep_path: Return result with original nesting
        dig_yaml_lists: Also dig into single-key->value pairs, as often found in yaml.
    """
    for i in sections:
        if isinstance(data, dict):
            if child := data.get(i):
                data = child
            else:
                return None
        elif dig_yaml_lists and isinstance(data, list):
            # this part is for yaml-style listitems
            for idx in data:
                if i in idx and isinstance(idx, dict):
                    data = idx[i]
                    break
                if isinstance(idx, str) and idx == i:
                    data = idx
                    break
            else:
                return None
    if not keep_path:
        return data
    result: dict[str, Any] = {}
    new = result
    for sect in sections:
        result[sect] = data if sect == sections[-1] else {}
        result = result[sect]
    return new


def merge(
    target: list | dict[str, Any],
    *source: list | dict[str, Any],
    deepcopy: bool = False,
    mergers: dict[type, Callable[[Any, Any, Any], Any]] | None = None,
) -> list | dict[str, Any]:
    """Merge given data structures using mergers provided.

    Args:
        target: Data structure to merge into
        source:  Data structures to merge into target
        deepcopy: Whether to deepcopy the target
        mergers: Mergers with strategies for each type (default: additive)
    """
    import copy

    if deepcopy:
        target = copy.deepcopy(target)
    context = deepmerge.DeepMerger(mergers)
    for s in source:
        target = context.merge(s, target)
    return target



"""Module for deep merging of data structures with customizable merge strategies."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any


if TYPE_CHECKING:
    from collections.abc import Callable, Mapping


def merge_dict(merger: DeepMerger, source: Mapping, target: Mapping) -> Mapping:
    """Merge two mappings recursively.

    Args:
        merger: The DeepMerger instance handling the merge operation
        source: The source mapping whose values take precedence
        target: The target mapping to merge into

    Returns:
        A new dictionary containing the merged key-value pairs

    Example:
        ```python
        merger = DeepMerger()
        source = {"a": {"b": 1}}
        target = {"a": {"c": 2}}
        merge_dict(merger, source, target)
        # {'a': {'b': 1, 'c': 2}}
        ```
    """
    result = dict(target)
    for key, source_value in source.items():
        target_value = result.get(key, type(source_value)())
        try:
            value = merger.merge(source_value, target_value)
        except TypeError:
            # If values can't be merged, use source value
            value = source_value
        result[key] = value
    return result


def merge_list(merger: DeepMerger, source: list, target: list) -> list:
    """Concatenate two lists.

    Args:
        merger: The DeepMerger instance handling the merge operation
        source: The source list to append
        target: The target list to merge into

    Returns:
        A new list containing all elements from both lists

    Example:
        ```python
        merger = DeepMerger()
        merge_list(merger, [3, 4], [1, 2])
        # [1, 2, 3, 4]
        ```
    """
    return target + source


DEFAULT_MERGERS: dict[type, Callable] = {
    dict: merge_dict,
    list: merge_list,
}


class DeepMerger:
    """A class that handles deep merging of data structures.

    The merger can be customized by providing different merge strategies
    for different types.

    Attributes:
        mergers: A dictionary mapping types to their corresponding merge functions

    Example:
        ```python
        merger = DeepMerger()
        source = {"a": {"b": 1}}
        target = {"a": {"c": 2}}
        merger.merge(source, target)
        # {'a': {'b': 1, 'c': 2}}
        ```
    """

    mergers: dict[type[Any], Callable[..., Any]] = DEFAULT_MERGERS

    def __init__(self, mergers: dict[type[Any], Callable[..., Any]] | None = None) -> None:
        """Initialize the DeepMerger with custom merge strategies.

        Args:
            mergers: Optional dictionary of type-specific merge functions
        """
        if mergers is not None:
            self.mergers = mergers

    def merge[T](self, source: T, target: T) -> T:
        """Merge two objects of the same type.

        Args:
            source: The source object whose values take precedence
            target: The target object to merge into

        Returns:
            The merged object

        Raises:
            TypeError: If the types cannot be merged
        """
        source_type = type(source)
        target_type = type(target)
        merger = self.mergers.get(target_type)
        if source_type is not target_type or merger is None:
            msg = f"Cannot merge {source_type} with {target_type}"
            raise TypeError(msg)
        return merger(self, source, target)


if __name__ == "__main__":
    merger = DeepMerger()
    source = {"a": {"b": 1}}
    target = {"a": {"c": 2}}
    result = merger.merge(source, target)
    print(result)
    assert result == {"a": {"b": 1, "c": 2}}



from __future__ import annotations

import dataclasses
import functools
import os
import tomllib
from typing import TYPE_CHECKING, Any, Literal, Self

import jinjarope
from jinjarope import envconfig, envglobals, envtests, loaders, utils


if TYPE_CHECKING:
    from collections.abc import Callable

    import jinja2


@functools.cache
def _load(txt: str) -> dict[str, Any]:
    """Some dumb caching since this might be invoked on each env instanciation."""
    return tomllib.loads(txt)


class JinjaFile(dict[str, Any]):
    """A file defining filters / tests."""

    def __init__(self, path: str | os.PathLike[str]) -> None:
        """Instanciate the file.

        Args:
            path: Path to the jinja file
        """
        super().__init__()
        text = envglobals.load_file_cached(os.fspath(path))
        data = _load(text)
        self.update(data)

    @property
    def filters(self) -> list[JinjaItem]:
        """Return list of filters defined in the file."""
        return [
            JinjaItem(filter_name, typ="filter", **dct)
            for filter_name, dct in self.get("filters", {}).items()
            if all(envtests.is_installed(i) for i in dct.get("required_packages", []))
        ]

    @property
    def tests(self) -> list[JinjaItem]:
        """Return list of tests defined in the file."""
        return [
            JinjaItem(filter_name, typ="test", **dct)
            for filter_name, dct in self.get("tests", {}).items()
            if all(envtests.is_installed(i) for i in dct.get("required_packages", []))
        ]

    @property
    def functions(self) -> list[JinjaItem]:
        """Return list of functions defined in the file."""
        return [
            JinjaItem(filter_name, typ="function", **dct)
            for filter_name, dct in self.get("functions", {}).items()
            if all(envtests.is_installed(i) for i in dct.get("required_packages", []))
        ]

    @property
    def filters_dict(self) -> dict[str, Callable[..., Any]]:
        """Return a dictionary with all filters.

        Can directly get merged into env filters.
        """
        dct = {}
        for f in self.filters:
            dct[f.identifier] = f.filter_fn
            for alias in f.aliases:
                dct[alias] = f.filter_fn
        return dct

    @property
    def tests_dict(self) -> dict[str, Callable[..., bool]]:
        """Return a dictionary with all filters.

        Can directly get merged into env filters.
        """
        dct = {}
        for f in self.tests:
            dct[f.identifier] = f.filter_fn
            for alias in f.aliases:
                dct[alias] = f.filter_fn
        return dct

    @property
    def functions_dict(self) -> dict[str, Callable[..., Any]]:
        """Return a dictionary with all filters.

        Can directly get merged into env filters.
        """
        dct = {}
        for f in self.functions:
            dct[f.identifier] = f.filter_fn
            for alias in f.aliases:
                dct[alias] = f.filter_fn
        return dct

    @property
    def envconfig(self) -> envconfig.EnvConfig:
        """Return the config object defined in this Jinja file."""
        cfg = self.get("config", {})
        return envconfig.EnvConfig(**cfg, loader=self.loader)

    @property
    def loader(self) -> jinja2.BaseLoader | None:
        """Return a (composed Choice-) loader defined in this Jinja file."""
        return loaders.from_json(self.get("loaders", []))


@dataclasses.dataclass(frozen=True)
class JinjaItem:
    """An item representing a filter / test."""

    identifier: str
    typ: Literal["filter", "test", "function"]
    fn: str
    group: str
    icon: str | None = None
    examples: dict[str, dict[str, str]] = dataclasses.field(default_factory=dict)
    description: str | None = None
    aliases: list[str] = dataclasses.field(default_factory=list)
    required_packages: list[str] = dataclasses.field(default_factory=list)

    def __repr__(self):
        return utils.get_repr(self, self.identifier)

    @property
    def filter_fn(self) -> Callable[..., Any]:
        """Return the callable to use as filter / test / function."""
        try:
            obj = utils.resolve(self.fn)
        except AttributeError:
            msg = f"Could not import jinja item {self.identifier!r} from {self.fn!r}"
            raise ImportError(msg) from AttributeError
        if not callable(obj):
            msg = "Filter needs correct, importable Path for callable"
            raise TypeError(msg)
        return obj

    @classmethod
    def for_function(
        cls,
        fn: Callable[..., Any],
        typ: Literal["filter", "test", "function"],
        group: str = "imported",
        **kwargs: Any,
    ) -> Self:
        """Alternative ctor to construct a JinjaItem based on a callable.

        Args:
            fn: Callable to get a JinjaItem for
            typ: The item type
            group: Group for metadata
            kwargs: Additional keyword arguments for JinjaItem ctor
        """
        return cls(
            fn.__name__,
            typ=typ,
            fn=f"{fn.__module__}.{fn.__name__}",
            group=group,
            **kwargs,
        )

    def apply(self, *args: Any, **kwargs: Any) -> Any:
        """Apply the filter function using given arguments and keywords.

        Args:
            args: The arguments for the call
            kwargs: They keyword arguments for the call
        """
        return self.filter_fn(*args, **kwargs)

    def resolve_example(self, example_name: str) -> str:
        """Render example with given name and return the result.

        Args:
            example_name: The example identifier
        """
        example = self.examples[example_name]
        loader = jinjarope.FileSystemLoader("")
        env = jinjarope.Environment(loader=loader)
        return env.render_string(example["template"])


if __name__ == "__main__":
    file = JinjaFile("src/jinjarope/resources/filters.toml")
    print(file.filters[5].resolve_example("basic"))



from __future__ import annotations

import base64
import functools
import logging
import posixpath
import re
from typing import TYPE_CHECKING, Any, Literal, TypeVar


if TYPE_CHECKING:
    from collections.abc import Mapping
    from xml.etree import ElementTree as ET


logger = logging.getLogger(__name__)

QueryStr = Literal[
    "fragment",
    "hostname",
    "netloc",
    "password",
    "path",
    "port",
    "query",
    "scheme",
    "username",
]

ANSI_STYLES = {
    1: {"font_weight": "bold"},
    2: {"font_weight": "lighter"},
    3: {"font_weight": "italic"},
    4: {"text_decoration": "underline"},
    5: {"text_decoration": "blink"},
    6: {"text_decoration": "blink"},
    8: {"visibility": "hidden"},
    9: {"text_decoration": "line-through"},
    30: {"color": "black"},
    31: {"color": "red"},
    32: {"color": "green"},
    33: {"color": "yellow"},
    34: {"color": "blue"},
    35: {"color": "magenta"},
    36: {"color": "cyan"},
    37: {"color": "white"},
}


def wrap_in_elem(text: str | None, tag: str, add_linebreaks: bool = False, **kwargs: Any) -> str:
    """Wrap given text in an HTML/XML tag (with attributes).

    If text is empty, just return an empty string.

    Args:
        text: Text to wrap
        tag: Tag to wrap text in
        add_linebreaks: Adds a linebreak before and after the text
        kwargs: additional key-value pairs to be inserted as attributes for tag.
                Key strings will have "_" stripped from the end to allow using keywords.
    """
    if not text:
        return ""
    attrs = [f'{k.rstrip("_")}="{v}"' for k, v in kwargs.items()]
    attr_str = (" " + " ".join(attrs)) if attrs else ""
    nl = "\n" if add_linebreaks else ""
    return f"<{tag}{attr_str}>{nl}{text}{nl}</{tag}>"


def html_link(text: str | None = None, link: str | None = None, **kwargs: Any) -> str:
    """Create a html link.

    If link is empty string or None, just the text will get returned.

    Args:
        text: Text to show for the link
        link: Target url
        kwargs: additional key-value pairs to be inserted as attributes.
                Key strings will have "_" stripped from the end to allow using keywords.
    """
    if not link:
        return text or ""
    attrs = [f'{k.rstrip("_")}="{v}"' for k, v in kwargs.items()]
    attr_str = (" " + " ".join(attrs)) if attrs else ""
    return f"<a href={link!r}{attr_str}>{text or link}</a>"


def format_js_map(mapping: dict[str, Any] | str, indent: int = 4) -> str:
    """Return JS map str for given dictionary.

    Args:
        mapping: Dictionary to dump
        indent: The amount of indentation for the key-value pairs
    """
    import anyenv

    dct = anyenv.load_json(mapping) if isinstance(mapping, str) else mapping
    rows: list[str] = []
    indent_str = " " * indent
    for k, v in dct.items():
        match v:
            case bool():
                rows.append(f"{indent_str}{k}: {str(v).lower()},")
            case dict():
                rows.append(f"{indent_str}{k}: {format_js_map(v)},")
            case None:
                rows.append(f"{indent_str}{k}: null,")
            case _:
                rows.append(f"{indent_str}{k}: {v!r},")
    row_str = "\n" + "\n".join(rows) + "\n"
    return f"{{{row_str}}}"


def svg_to_data_uri(svg: str) -> str:
    """Wrap svg as data URL.

    Args:
        svg: The svg to wrap into a data URL
    """
    if not isinstance(svg, str):
        msg = "Invalid type: %r"
        raise TypeError(msg, type(svg))
    return f"url('data:image/svg+xml;charset=utf-8,{svg}')"


def clean_svg(text: str) -> str:
    """Strip off unwanted stuff from svg text which might be added by external libs.

    Removes xml headers and doctype declarations.

    Args:
        text: The text to cleanup / filter
    """
    text = re.sub(r"<\?xml version.*\?>\s*", "", text, flags=re.DOTALL)
    text = re.sub(r"<!DOCTYPE svg.*?>", "", text, flags=re.DOTALL)
    return text.strip()


def format_css_rule(dct: Mapping) -> str:
    """Format a nested dictionary as CSS rule.

    Mapping must be of shape {".a": {"b": "c"}}

    Args:
        dct: The mapping to convert to CSS text
    """
    data: dict[str, list] = {}

    def _parse(obj, selector: str = "") -> None:
        for key, value in obj.items():
            if hasattr(value, "items"):
                rule = selector + " " + key
                data[rule] = []
                _parse(value, rule)

            else:
                prop = data[selector]
                prop.append(f"\t{key}: {value};\n")

    _parse(dct)
    string = ""
    for key, value in sorted(data.items()):
        if data[key]:
            string += key[1:] + " {\n" + "".join(value) + "}\n\n"
    return string


@functools.lru_cache
def format_xml(
    str_or_elem: str | ET.Element,
    indent: str | int = "  ",
    level: int = 0,
    method: Literal["xml", "html", "text", "c14n"] = "html",
    short_empty_elements: bool = True,
    add_declaration: bool = False,
) -> str:
    """(Pretty)print given XML.

    Args:
        str_or_elem: XML to prettyprint
        indent: Amount of spaces to use for indentation
        level: Initial indentation level
        method: Output method
        short_empty_elements: Whether empty elements should get printed in short form
                              (applies when mode is "xml")
        add_declaration: whether a XML declaration should be printed
                         (applies when mode is "xml")
    """
    from xml.etree import ElementTree as ET

    if isinstance(str_or_elem, str):
        str_or_elem = ET.fromstring(str_or_elem)
    space = indent if isinstance(indent, str) else indent * " "
    ET.indent(str_or_elem, space=space, level=level)
    return ET.tostring(
        str_or_elem,
        encoding="unicode",
        method=method,
        xml_declaration=add_declaration,
        short_empty_elements=short_empty_elements,
    )


def ansi2html(ansi_string: str, styles: dict[int, dict[str, str]] | None = None) -> str:
    """Convert ansi string to colored HTML.

    Args:
        ansi_string: text with ANSI color codes.
        styles: A mapping from ANSI codes to a dict with css

    Returns:
        HTML string
    """
    styles = styles or ANSI_STYLES
    previous_end = 0
    in_span = False
    ansi_codes: list[int] = []
    ansi_finder = re.compile("\033\\[([\\d;]*)([a-zA-z])")
    parts = []
    for match in ansi_finder.finditer(ansi_string):
        parts.append(ansi_string[previous_end : match.start()])
        previous_end = match.end()
        params, command = match.groups()

        if command not in "mM":
            continue

        try:
            params = [int(p) for p in params.split(";")]
        except ValueError:
            params = [0]

        for i, v in enumerate(params):
            if v == 0:
                params = params[i + 1 :]
                if in_span:
                    in_span = False
                    parts.append("</span>")
                ansi_codes = []
                if not params:
                    continue

        ansi_codes.extend(params)
        if in_span:
            parts.append("</span>")
            in_span = False

        if not ansi_codes:
            continue

        style = [
            "; ".join([f"{k}: {v}" for k, v in styles[k].items()]).strip()
            for k in ansi_codes
            if k in styles
        ]
        parts.append(f'<span style="{"; ".join(style)}">')

        in_span = True

    parts.append(ansi_string[previous_end:])
    if in_span:
        parts.append("</span>")
        in_span = False
    return "".join(parts)


def split_url(value: str, query: QueryStr | None = None) -> str | dict[str, str]:
    """Split a URL into its parts (and optionally return a specific part).

    Args:
        value: The URL to split
        query: Optional URL part to extract
    """
    from urllib.parse import urlsplit

    def object_to_dict(obj: Any, exclude: list[str] | None = None) -> dict[str, Any]:
        """Converts an object into a dict making the properties into keys.

        Allows excluding certain keys.
        """
        if exclude is None or not isinstance(exclude, list):
            exclude = []
        return {
            key: getattr(obj, key)
            for key in dir(obj)
            if not (key.startswith("_") or key in exclude)
        }

    to_exclude = ["count", "index", "geturl", "encode"]
    results = object_to_dict(urlsplit(value), exclude=to_exclude)

    # If a query is supplied, make sure it's valid then return the results.
    # If no option is supplied, return the entire dictionary.
    if not query:
        return results
    if query not in results:
        msg = "split_url: unknown URL component: %s"
        raise ValueError(msg, query)
    return results[query]


@functools.lru_cache
def _get_norm_url(path: str) -> tuple[str, int]:
    from urllib.parse import urlsplit

    if not path:
        path = "."
    elif "\\" in path:
        logger.warning(
            "Path %r uses OS-specific separator '\\'. "
            "That will be unsupported in a future release. Please change it to '/'.",
            path,
        )
        path = path.replace("\\", "/")
    # Allow links to be fully qualified URLs
    parsed = urlsplit(path)
    if parsed.scheme or parsed.netloc or path.startswith(("/", "#")):
        return path, -1

    # Relative path - preserve information about it
    norm = posixpath.normpath(path) + "/"
    relative_level = 0
    while norm.startswith("../", relative_level * 3):
        relative_level += 1
    return path, relative_level


def normalize_url(path: str, url: str | None = None, base: str = "") -> str:
    """Return a URL relative to the given url or using the base.

    Args:
        path: The path to normalize
        url: Optional relative url
        base: Base path
    """
    path, relative_level = _get_norm_url(path)
    if relative_level == -1:
        return path
    if url is None:
        return posixpath.join(base, path)
    result = relative_url(url, path)
    if relative_level > 0:
        result = "../" * relative_level + result
    return result


@functools.cache
def _norm_parts(path: str) -> list[str]:
    if not path.startswith("/"):
        path = "/" + path
    path = posixpath.normpath(path)[1:]
    return path.split("/") if path else []


def relative_url_mkdocs(url: str, other: str) -> str:
    """Return given url relative to other (MkDocs implementation).

    Both are operated as slash-separated paths, similarly to the 'path' part of a URL.
    The last component of `other` is skipped if it contains a dot (considered a file).
    Actual URLs (with schemas etc.) aren't supported. The leading slash is ignored.
    Paths are normalized ('..' works as parent directory), but going higher than the
    root has no effect ('foo/../../bar' ends up just as 'bar').

    Args:
        url: URL A.
        other: URL B.
    """
    # Remove filename from other url if it has one.
    dirname, _, basename = other.rpartition("/")
    if "." in basename:
        other = dirname

    other_parts = _norm_parts(other)
    dest_parts = _norm_parts(url)
    common = 0
    for a, b in zip(other_parts, dest_parts):
        if a != b:
            break
        common += 1

    rel_parts = [".."] * (len(other_parts) - common) + dest_parts[common:]
    relurl = "/".join(rel_parts) or "."
    return relurl + "/" if url.endswith("/") else relurl


def relative_url(url_a: str, url_b: str) -> str:
    """Compute the relative path from URL A to URL B.

    Args:
        url_a: URL A.
        url_b: URL B.

    Returns:
        The relative URL to go from A to B.
    """
    parts_a = url_a.split("/")
    if "#" in url_b:
        url_b, anchor = url_b.split("#", 1)
    else:
        anchor = None
    parts_b = url_b.split("/")

    # remove common left parts
    while parts_a and parts_b and parts_a[0] == parts_b[0]:
        parts_a.pop(0)
        parts_b.pop(0)

    # go up as many times as remaining a parts' depth
    levels = len(parts_a) - 1
    parts_relative = [".."] * levels + parts_b
    relative = "/".join(parts_relative)
    return f"{relative}#{anchor}" if anchor else relative


def url_to_b64(image_url: str) -> str | None:
    """Convert an image URL to a base64-encoded string.

    Args:
        image_url: The URL of the image to convert.

    Returns:
        The base64-encoded string of the image.

    Raises:
        requests.RequestException: If there's an error downloading the image.
    """
    import httpx

    # Download the image
    with httpx.Client(follow_redirects=True) as client:
        response = client.get(image_url)
        response.raise_for_status()
        image_data = response.content

    # Encode the image to base64
    return base64.b64encode(image_data).decode("utf-8")


StrOrBytes = TypeVar("StrOrBytes", bytes, str)


# TypeVar for maintaining input/output type consistency
ContentType = TypeVar("ContentType", str, bytes)
Position = Literal["body", "head", "end_head", "end_body"]


def inject_javascript[ContentType: (str, bytes)](
    html_content: ContentType,
    javascript: str,
    *,
    position: Position = "end_body",
) -> ContentType:
    """Injects JavaScript code into an HTML string or bytes object.

    Args:
        html_content: The HTML content to inject the JavaScript into
        javascript: The JavaScript code to inject
        position: The position to inject the JavaScript ('body' by default)

    Returns:
        The modified HTML content with the same type as the input

    Raises:
        ValueError: If the specified position tag is not found in the HTML content
        TypeError: If the input type is neither str nor bytes
    """
    # Convert bytes to str if needed
    is_bytes = isinstance(html_content, bytes)
    working_content: str = html_content.decode() if is_bytes else html_content  # type: ignore[assignment, attr-defined]

    # Prepare the JavaScript tag
    script_tag = f"<script>{javascript}</script>"

    # Define the injection patterns
    patterns = {
        "body": (r"<body[^>]*>", lambda m: f"{m.group(0)}{script_tag}"),
        "head": (r"<head[^>]*>", lambda m: f"{m.group(0)}{script_tag}"),
        "end_head": (r"</head>", lambda m: f"{script_tag}{m.group(0)}"),
        "end_body": (r"</body>", lambda m: f"{script_tag}{m.group(0)}"),
    }

    if position not in patterns:
        msg = f"Invalid position: {position}. Must be one of {list(patterns.keys())}"
        raise ValueError(msg)

    pattern, replacement = patterns[position]
    modified_content = re.sub(pattern, replacement, working_content, count=1)

    # If no substitution was made, the tag wasn't found
    if modified_content == working_content:
        msg = f"Could not find {position} tag in HTML content"
        raise ValueError(msg)
    # Return the same type as input
    return modified_content.encode() if is_bytes else modified_content  # type: ignore[return-value]


if __name__ == "__main__":
    print(format_js_map({"key": {"nested_key": "nested_value"}}))



from __future__ import annotations

import functools
import pathlib


PYCONIFY_TO_PREFIXES = {
    "mdi": "material",
    "simple-icons": "simple",
    "octicon": "octicons",
    "fa6-regular": "fontawesome-regular",
    "fa-brands": "fontawesome-brands",
    "fa6-solid": "fontawesome-solid",
}


ROOT = pathlib.Path(__file__).parent
ICON_FILE = ROOT / "resources" / "icons.json.gzip"


@functools.cache
def _get_collection_map(*prefixes: str) -> dict[str, list[str]]:
    """Return a dictionary with a mapping from pyconify name to icon prefixes.

    In order to provide compatibility with the materialx-icon-index,
    we also add the prefixes used by that index, which is different from
    the pyconify prefixes. (material vs mdi etc, see PYCONIFY_TO_PREFIXES)

    Args:
        prefixes: The collections to fetch
    """
    import pyconify

    mapping = {coll: [coll] for coll in pyconify.collections(*prefixes)}
    for k, v in PYCONIFY_TO_PREFIXES.items():
        if k in mapping:
            mapping[k].append(v)
    return mapping


@functools.cache
def _get_pyconify_icon_index(*collections: str) -> dict[str, dict[str, str]]:
    """Return a icon index for the pymdownx emoji extension containing pyconify icons.

    The dictionaries contain three key-value pairs:
    - "name": the emoji identifier
    - "path": the pyconify key
    - "set": the collection name

    Args:
        collections: Collections to fetch. If none given, fetch all
    """
    import pyconify

    index = {}
    for coll, prefixes in _get_collection_map(*collections).items():
        collection = pyconify.collection(coll)
        for icon_name in collection.get("uncategorized", []):
            for prefix in prefixes:
                name = f":{prefix}-{icon_name}:"
                index[name] = {
                    "name": name,
                    "path": f"{coll}:{icon_name}",
                    "set": coll,
                }
        for cat in pyconify.collection(coll).get("categories", {}).values():
            for icon_name in cat:
                for prefix in prefixes:
                    name = f":{prefix}-{icon_name}:"
                    index[name] = {
                        "name": name,
                        "path": f"{coll}:{icon_name}",
                        "set": coll,
                    }
    return index


def write_icon_index() -> None:
    """Fetch the complete icon index and write it gzipped to disk."""
    import gzip
    import json

    mapping = _get_pyconify_icon_index()
    with gzip.open(ICON_FILE, "w") as file:
        file.write(json.dumps(mapping).encode())


def load_icon_index() -> dict[str, dict[str, str]]:
    """Load the complete icon index from disk."""
    import gzip

    import anyenv

    with gzip.open(ICON_FILE, "r") as file:
        return anyenv.load_json(file.read())


if __name__ == "__main__":
    # idx = load_icon_index()
    # print(idx)
    write_icon_index()



from __future__ import annotations

from typing import Literal

import jinja2


UndefinedStr = Literal["keep", "silent", "strict", "lax"]


class LaxUndefined(jinja2.Undefined):
    """Pass anything wrong as blank."""

    def _fail_with_undefined_error(self, *args, **kwargs) -> str:  # type: ignore[override]
        return ""


UNDEFINED_BEHAVIOR: dict[UndefinedStr, type[jinja2.Undefined]] = {
    "keep": jinja2.DebugUndefined,
    "silent": jinja2.Undefined,
    "strict": jinja2.StrictUndefined,
    # lax will even pass unknown objects:
    "lax": LaxUndefined,
}


if __name__ == "__main__":
    pass



from __future__ import annotations

import importlib


class LazyLiteLLM:
    _lazy_module = None

    def __getattr__(self, name):
        if name == "_lazy_module":
            return super()
        self._load_litellm()
        return getattr(self._lazy_module, name)

    def _load_litellm(self) -> None:
        if self._lazy_module is not None:
            return

        self._lazy_module = importlib.import_module("litellm")

        self._lazy_module.suppress_debug_info = True  # type: ignore[attr-defined]
        self._lazy_module.set_verbose = False  # type: ignore[attr-defined]
        self._lazy_module.drop_params = True  # type: ignore[attr-defined]
        self._lazy_module._logging._disable_debugging()


if __name__ == "__main__":
    litellm = LazyLiteLLM()



from __future__ import annotations

from collections.abc import Callable, Iterator
import contextlib
import functools
import inspect
import logging
import pathlib
import types
from typing import Any


logger = logging.getLogger(__name__)


HasCodeType = (
    types.ModuleType
    | type
    | types.MethodType
    | types.FunctionType
    | types.TracebackType
    | types.FrameType
    | types.CodeType
    | Callable[..., Any]
)


@functools.cache
def list_subclasses[ClassType: type](
    klass: ClassType,
    *,
    recursive: bool = True,
    filter_abstract: bool = False,
    filter_generic: bool = True,
    filter_locals: bool = True,
) -> list[ClassType]:
    """Return list of all subclasses of given klass.

    Note: This call is cached. Consider iter_subclasses for uncached iterating.

    Args:
        klass: class to get subclasses from
        filter_abstract: whether abstract base classes should be included.
        filter_generic: whether generic base classes should be included.
        filter_locals: whether local base classes should be included.
        recursive: whether to also get subclasses of subclasses.
    """
    return list(
        iter_subclasses(
            klass,
            recursive=recursive,
            filter_abstract=filter_abstract,
            filter_generic=filter_generic,
            filter_locals=filter_locals,
        ),
    )


def iter_subclasses[ClassType: type](
    klass: ClassType,
    *,
    recursive: bool = True,
    filter_abstract: bool = False,
    filter_generic: bool = True,
    filter_locals: bool = True,
) -> Iterator[ClassType]:
    """(Recursively) iterate all subclasses of given klass.

    Args:
        klass: class to get subclasses from
        filter_abstract: whether abstract base classes should be included.
        filter_generic: whether generic base classes should be included.
        filter_locals: whether local base classes should be included.
        recursive: whether to also get subclasses of subclasses.
    """
    if getattr(klass.__subclasses__, "__self__", None) is None:
        return
    for kls in klass.__subclasses__():
        if recursive:
            yield from iter_subclasses(
                kls,
                filter_abstract=filter_abstract,
                filter_generic=filter_generic,
                filter_locals=filter_locals,
            )
        if filter_abstract and inspect.isabstract(kls):
            continue
        if filter_generic and kls.__qualname__.endswith("]"):
            continue
        if filter_locals and "<locals>" in kls.__qualname__:
            continue
        yield kls


@functools.cache
def list_baseclasses(
    klass: type,
    *,
    recursive: bool = True,
    filter_abstract: bool = False,
    filter_generic: bool = True,
    filter_locals: bool = True,
) -> list[type]:
    """Return list of all baseclasses of given klass.

    Args:
        klass: class to get subclasses from
        filter_abstract: whether abstract base classes should be included.
        filter_generic: whether generic base classes should be included.
        filter_locals: whether local base classes should be included.
        recursive: whether to also get baseclasses of baseclasses.
    """
    return list(
        iter_baseclasses(
            klass,
            recursive=recursive,
            filter_abstract=filter_abstract,
            filter_generic=filter_generic,
            filter_locals=filter_locals,
        ),
    )


def iter_baseclasses(
    klass: type,
    *,
    recursive: bool = True,
    filter_abstract: bool = False,
    filter_generic: bool = True,
    filter_locals: bool = True,
) -> Iterator[type]:
    """(Recursively) iterate all baseclasses of given klass.

    Args:
        klass: class to get subclasses from
        filter_abstract: whether abstract base classes should be included.
        filter_generic: whether generic base classes should be included.
        filter_locals: whether local base classes should be included.
        recursive: whether to also get baseclasses of baseclasses.
    """
    for kls in klass.__bases__:
        if recursive:
            yield from iter_baseclasses(
                kls,
                recursive=recursive,
                filter_abstract=filter_abstract,
                filter_generic=filter_generic,
                filter_locals=filter_locals,
            )
        if filter_abstract and inspect.isabstract(kls):
            continue
        if filter_generic and kls.__qualname__.endswith("]"):
            continue
        if filter_locals and "<locals>" in kls.__qualname__:
            continue
        yield kls


@functools.cache
def get_doc(
    obj: Any,
    *,
    escape: bool = False,
    fallback: str = "",
    from_base_classes: bool = False,
    only_summary: bool = False,
    only_description: bool = False,
) -> str:
    """Get __doc__ for given object.

    Args:
        obj: Object to get docstrings from
        escape: Whether docstrings should get escaped
        fallback: Fallback in case docstrings dont exist
        from_base_classes: Use base class docstrings if docstrings dont exist
        only_summary: Only return first line of docstrings
        only_description: Only return block after first line
    """
    from jinjarope import mdfilters

    match obj:
        case _ if from_base_classes:
            doc = inspect.getdoc(obj)
        case _ if obj.__doc__:
            doc = inspect.cleandoc(obj.__doc__)
        case _:
            doc = None
    if not doc:
        return fallback
    if only_summary:
        doc = doc.split("\n")[0]
    if only_description:
        doc = "\n".join(doc.split("\n")[1:])
    return mdfilters.md_escape(doc) if doc and escape else doc


def get_argspec(obj: Any, remove_self: bool = True) -> inspect.FullArgSpec:
    """Return a cleaned-up FullArgSpec for given callable.

    ArgSpec is cleaned up by removing `self` from method callables.

    Args:
        obj: A callable python object
        remove_self: Whether to remove "self" argument from method argspecs
    """
    if inspect.isfunction(obj):
        argspec = inspect.getfullargspec(obj)
    elif inspect.ismethod(obj):
        argspec = inspect.getfullargspec(obj)
        if remove_self:
            del argspec.args[0]
    elif inspect.isclass(obj):
        if obj.__init__ is object.__init__:  # to avoid an error
            argspec = inspect.getfullargspec(lambda self: None)
        else:
            argspec = inspect.getfullargspec(obj.__init__)
        if remove_self:
            del argspec.args[0]
    elif callable(obj):
        argspec = inspect.getfullargspec(obj.__call__)
        if remove_self:
            del argspec.args[0]
    else:
        msg = f"{obj} is not callable"
        raise TypeError(msg)
    return argspec


def get_deprecated_message(obj: Any) -> str | None:
    """Return deprecated message (created by deprecated decorator).

    Args:
        obj: Object to check
    """
    return obj.__deprecated__ if hasattr(obj, "__deprecated__") else None


@functools.cache
def get_source(obj: HasCodeType) -> str:
    """Cached wrapper for inspect.getsource.

    Args:
        obj: Object to return source for.
    """
    return inspect.getsource(obj)


@functools.cache
def get_source_lines(obj: HasCodeType) -> tuple[list[str], int]:
    """Cached wrapper for inspect.getsourcelines.

    Args:
        obj: Object to return source lines for.
    """
    return inspect.getsourcelines(obj)


@functools.cache
def get_signature(obj: Callable[..., Any]) -> inspect.Signature:
    """Cached wrapper for inspect.signature.

    Args:
        obj: Callable to get a signature for.
    """
    return inspect.signature(obj)


@functools.cache
def get_members(obj: object, predicate: Callable[[Any], bool] | None = None):
    """Cached version of inspect.getmembers.

    Args:
        obj: Object to get members for
        predicate: Optional predicate for the members
    """
    return inspect.getmembers(obj, predicate)


@functools.cache
def get_file(obj: HasCodeType) -> pathlib.Path | None:
    """Cached wrapper for inspect.getfile.

    Args:
        obj: Object to get file for
    """
    with contextlib.suppress(TypeError):
        return pathlib.Path(inspect.getfile(obj))
    return None


if __name__ == "__main__":
    doc = get_doc(str)



"""Jinja2 template tag extensions."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Literal

from jinja2 import nodes
from jinja2.ext import Extension
from jinja2.lexer import describe_token


if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

    from jinja2.environment import Environment
    from jinja2.parser import Parser
    from jinja2.runtime import Context


__all__ = ["ContainerTag", "InclusionTag", "StandaloneTag"]
__version__ = "0.6.1"


class BaseTemplateTag(Extension):
    """Base class for template tag extensions providing common functionality."""

    def __init__(self, environment: Environment) -> None:
        super().__init__(environment)
        self.context: Context | None = None
        self.template: str | None = None
        self.lineno: int | None = None
        self.tag_name: str | None = None

    def parse(self, parser: Parser) -> nodes.Node:
        """Parse the tag definition and return a Node."""
        lineno = parser.stream.current.lineno
        tag_name = parser.stream.current.value
        meta_kwargs = [
            nodes.Keyword("_context", nodes.ContextReference()),
            nodes.Keyword("_template", nodes.Const(parser.name)),
            nodes.Keyword("_lineno", nodes.Const(lineno)),
            nodes.Keyword("_tag_name", nodes.Const(tag_name)),
        ]

        self.init_parser(parser)
        args, kwargs, options = self.parse_args(parser)
        kwargs.extend(meta_kwargs)
        options["tag_name"] = tag_name
        return self.create_node(parser, args, kwargs, lineno=lineno, **options)

    def init_parser(self, parser: Parser) -> None:
        """Initialize parser by skipping the tag name."""
        parser.stream.skip(1)

    def parse_args(
        self,
        parser: Parser,
    ) -> tuple[list[nodes.Expr], list[nodes.Keyword], dict[str, Any]]:
        """Parse arguments from the tag definition."""
        args: list[nodes.Expr] = []
        kwargs: list[nodes.Keyword] = []
        options: dict[str, str | None] = {"target": None}
        require_comma = False
        arguments_finished = False

        while parser.stream.current.type != "block_end":
            if parser.stream.current.test("name:as"):
                parser.stream.skip(1)
                options["target"] = parser.stream.expect("name").value
                arguments_finished = True

            if arguments_finished:
                if not parser.stream.current.test("block_end"):
                    desc = describe_token(parser.stream.current)
                    parser.fail(
                        f"Expected 'block_end', got {desc!r}",
                        parser.stream.current.lineno,
                    )
                break

            if require_comma:
                parser.stream.expect("comma")

                # support for trailing comma
                if parser.stream.current.type == "block_end":
                    break

            token = parser.stream.current
            if token.type == "name" and parser.stream.look().type == "assign":
                key = token.value
                parser.stream.skip(2)  # Skip name and assign tokens
                value = parser.parse_expression()
                kwargs.append(nodes.Keyword(key, value, lineno=value.lineno))
            else:
                if kwargs:
                    parser.fail("Invalid argument syntax", token.lineno)
                args.append(parser.parse_expression())

            require_comma = True

        return args, kwargs, options

    def create_node(
        self,
        parser: Parser,
        args: list[nodes.Expr],
        kwargs: list[nodes.Keyword],
        *,
        lineno: int,
        **options: Any,
    ) -> nodes.Node:
        raise NotImplementedError


class StandaloneTag(BaseTemplateTag):
    """Tag that renders to a single output without content block."""

    safe_output: ClassVar[bool] = False

    def create_node(
        self,
        parser: Parser,
        args: list[nodes.Expr],
        kwargs: list[nodes.Keyword],
        *,
        lineno: int,
        **options: Any,
    ) -> nodes.Node:
        call_node: nodes.Call | nodes.MarkSafeIfAutoescape = self.call_method(
            "render_wrapper",
            args,
            kwargs,
            lineno=lineno,
        )
        if self.safe_output:
            call_node = nodes.MarkSafeIfAutoescape(call_node, lineno=lineno)

        if target := options.get("target"):
            target_node = nodes.Name(target, "store", lineno=lineno)
            return nodes.Assign(target_node, call_node, lineno=lineno)

        return nodes.Output([call_node], lineno=lineno)

    def render_wrapper(self, *args: Any, **kwargs: Any) -> Any:
        self.context = kwargs.pop("_context")
        self.template = kwargs.pop("_template")
        self.lineno = kwargs.pop("_lineno")
        self.tag_name = kwargs.pop("_tag_name")
        return self.render(*args, **kwargs)

    def render(self, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError


class ContainerTag(BaseTemplateTag):
    """Tag that wraps content and processes it."""

    def create_node(
        self,
        parser: Parser,
        args: list[nodes.Expr],
        kwargs: list[nodes.Keyword],
        *,
        lineno: int,
        **options: Any,
    ) -> nodes.Node:
        """Create a node that processes wrapped content."""
        call_node = self.call_method("render_wrapper", args, kwargs, lineno=lineno)
        body = parser.parse_statements(
            (f"name:end{options['tag_name']}",),
            drop_needle=True,
        )
        call_block = nodes.CallBlock(call_node, [], [], body).set_lineno(lineno)

        if target := options.get("target"):
            target_node = nodes.Name(target, "store", lineno=lineno)
            return nodes.AssignBlock(target_node, None, [call_block], lineno=lineno)
        return call_block

    def render_wrapper(self, *args: Any, **kwargs: Any) -> Any:
        self.context = kwargs.pop("_context")
        self.template = kwargs.pop("_template")
        self.lineno = kwargs.pop("_lineno")
        self.tag_name = kwargs.pop("_tag_name")
        return self.render(*args, **kwargs)

    def render(self, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError


class InclusionTag(StandaloneTag):
    """Tag that includes other templates with context."""

    template_name: str | None = None
    safe_output: ClassVar[bool] = True

    def render(self, *args: Any, **kwargs: Any) -> str:
        """Render included template with context."""
        template_names = self.get_template_names(*args, **kwargs)
        template = (
            self.environment.get_template(template_names)
            if isinstance(template_names, str)
            else self.environment.select_template(template_names)
        )

        if not self.context:
            msg = "Context not available"
            raise RuntimeError(msg)

        context = template.new_context(
            {**self.context.get_all(), **self.get_context(*args, **kwargs)},
            shared=True,
        )
        return template.render(context)

    def get_context(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Get additional context for template."""
        return {}

    def get_template_names(self, *args: Any, **kwargs: Any) -> str | Sequence[str]:
        """Get template name(s) to include."""
        if not self.template_name:
            msg = "InclusionTag requires 'template_name' or 'get_template_names()'"
            raise RuntimeError(msg)
        return self.template_name


def create_tag_extension(
    typ: Literal["standalone", "container", "inclusion"],
    tag: str | list[str] | set[str],
    render_fn: Callable[..., str],
):
    """Create a Jinja2 extension from a render function."""
    jinja_tag = {tag} if isinstance(tag, str) else set(tag)
    match typ:
        case "standalone":
            base_cls: type[BaseTemplateTag] = StandaloneTag
        case "container":
            base_cls = ContainerTag
        case "inclusion":
            base_cls = InclusionTag
        case _:
            msg = f"Invalid extension type: {typ!r}"
            raise ValueError(msg)

    class Extension(base_cls):  # type: ignore
        tags = jinja_tag
        render = staticmethod(render_fn)

    return Extension


if __name__ == "__main__":
    import hmac

    import jinja2

    def render(secret, digest="sha256", caller=None):
        content = str(caller()).encode()

        if isinstance(secret, str):
            secret = secret.encode()

        signing = hmac.new(secret, content, digestmod=digest)
        return signing.hexdigest()

    HMACExtension = create_tag_extension("container", "hmac", render)

    env = jinja2.Environment(extensions=[HMACExtension])
    template = env.from_string("{% hmac 'SECRET', digest='sha1' %}Hello world!{% endhmac %}")
    print(template.render())
